/* (c) 2000, 2001 Michael Hudson
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include <icq.h>
#include "Python.h"
#include "ExtensionClass.h"

/* yawn... */

static PyObject* name_Logged;
static PyObject* name_Disconnected;
static PyObject* name_RecvMessage;
static PyObject* name_RecvURL;
static PyObject* name_RecvWebPager;
static PyObject* name_RecvMailExpress;
static PyObject* name_RecvChatReq;
static PyObject* name_RecvFileReq;
static PyObject* name_RecvAdded;
static PyObject* name_RecvAuthReq;
static PyObject* name_UserFound;
static PyObject* name_SearchDone;
static PyObject* name_UserOnline;
static PyObject* name_UserOffline;
static PyObject* name_UserStatusUpdate;
static PyObject* name_InfoReply;
static PyObject* name_ExtInfoReply;
static PyObject* name_WrongPassword;
static PyObject* name_InvalidUIN;
static PyObject* name_Log;
static PyObject* name_SrvAck;
static PyObject* name_RequestNotify;
static PyObject* name_FileNotify;
static PyObject* name_ChatNotify;
static PyObject* name_NewUIN;
static PyObject* name_MetaUserFound;
static PyObject* name_MetaUserInfo;
static PyObject* name_MetaUserWork;
static PyObject* name_MetaUserMore;
static PyObject* name_MetaUserAbout;
static PyObject* name_MetaUserInterests;
static PyObject* name_MetaUserAffiliations;
static PyObject* name_MetaUserHomePageCategory;

static PyObject* name_Uin;
#define i_Uin 0
static PyObject* name_OurIP;
#define i_OurIP 1
static PyObject* name_OurPort;
#define i_OurPort 2
static PyObject* name_ContactList;
#define i_ContactList 3
static PyObject* name_Status;
#define i_Status 4
static PyObject* name_Password;
#define i_Password 5
static PyObject* name_Nick;
#define i_Nick 6

static PyObject* name_to_i;

/*staticforward PyExtensionClass PyICQLinkType;*/

typedef struct {
	PyObject_HEAD

	icq_Link* link;
	PyObject* chat_sessions;
	PyObject* file_sessions;
} PyICQLink;

staticforward PyTypeObject PyICQFileSessionType;

typedef struct {
	PyObject_HEAD
	
	icq_FileSession* fs;
} PyICQFileSession;

staticforward PyTypeObject PyICQChatSessionType;

typedef struct {
	PyObject_HEAD
	
	icq_ChatSession* cs;
} PyICQChatSession;

static int boolconv(PyObject* obj, void* p)
{
	*(char*)p = PyObject_IsTrue(obj);
	return 1;
}

static PyObject* PyICQChatSession_New(icq_ChatSession* pchat);
static PyObject* PyICQFileSession_New(icq_FileSession* pfile);

/***********************************************************
 *                        callbacks                        *
 ***********************************************************/

static void
callback_helper(icq_Link* link, PyObject* name, char* format, ...)
{
	PyObject* ob;
	PyObject* callable;
	PyObject* tupleargs;
	PyObject* result;
	va_list va;

/*	printf("callback %s\n",PyString_AsString(name));*/

	ob = (PyObject*)link->icq_UserData;
	callable = PyObject_GetAttr(ob, name);
	va_start(va, format);
	tupleargs = Py_VaBuildValue(format, va);
	va_end(va);
	result = PyEval_CallObject(callable, tupleargs);
	Py_DECREF(tupleargs);
	Py_XDECREF(result);
}

static void 
icqlib_callback_Logged(icq_Link* link)
{
	callback_helper(link,name_Logged,"()");
}

static void 
icqlib_callback_Disconnected(icq_Link* link)
{
	callback_helper(link,name_Disconnected,"()");
}

static void 
icqlib_callback_RecvMessage(icq_Link* link, unsigned long uin,
       unsigned char hour, unsigned char minute, unsigned char day,
       unsigned char month, unsigned short year, const char *msg)
{
	callback_helper(link,name_RecvMessage,"(l(bbbbh)s)",
			uin,hour,minute,day,month,year,msg);
}

static void 
icqlib_callback_RecvURL(icq_Link* link, unsigned long uin,
       unsigned char hour, unsigned char minute, unsigned char day,
       unsigned char month, unsigned short year, const char *url,
       const char *descr)
{
	callback_helper(link,name_RecvURL,"(l(bbbbh)ss)",
			uin,hour,minute,day,month,year,url,descr);
}

static void 
icqlib_callback_RecvWebPager(icq_Link* link, unsigned char hour,
       unsigned char minute, unsigned char day, unsigned char month,
       unsigned short year, const char *nick, const char *email,
       const char *msg)
{
	callback_helper(link,name_RecvWebPager,"((bbbbh)sss)",
			hour,minute,day,month,year,nick,email,msg);
}

static void 
icqlib_callback_RecvMailExpress(icq_Link* link, unsigned char hour,
       unsigned char minute, unsigned char day, unsigned char month,
       unsigned short year, const char *nick, const char *email,
       const char *msg)
{
	callback_helper(link,name_RecvMailExpress,"((bbbbh)sss)",
			hour,minute,day,month,year,nick,email,msg);
}

static void 
icqlib_callback_RecvChatReq(icq_Link* link, unsigned long uin,
       unsigned char hour, unsigned char minute, unsigned char day,
       unsigned char month, unsigned short year, const char *descr,
       unsigned long seq)
{
	callback_helper(link,name_RecvChatReq,"(l(bbbbh)sl)",
			uin,hour,minute,day,month,year,descr,seq);
}

static void 
icqlib_callback_RecvFileReq(icq_Link* link, unsigned long uin,
       unsigned char hour, unsigned char minute, unsigned char day,
       unsigned char month, unsigned short year, const char *descr,
       const char *filename, unsigned long filesize, unsigned long seq)
{
	callback_helper(link,name_RecvFileReq,"(l(bbbbh)ssll)",
			uin,hour,minute,day,month,year,descr,filename,seq);
}

static void 
icqlib_callback_RecvAdded(icq_Link* link, unsigned long uin,
       unsigned char hour, unsigned char minute, unsigned char day,
       unsigned char month, unsigned short year, const char *nick,
       const char *first, const char *last, const char *email)
{
	callback_helper(link,name_RecvAdded,"(l(bbbbh)ssss)",
			uin,hour,minute,day,month,year,nick,first,last,email);
}

static void 
icqlib_callback_RecvAuthReq(icq_Link* link, unsigned long uin,
       unsigned char hour, unsigned char minute, unsigned char day,
       unsigned char month, unsigned short year, const char *nick,
       const char *first, const char *last, const char *email,
       const char *reason)
{
	callback_helper(link,name_RecvAuthReq,"(l(bbbbh)sssss)",
			uin,hour,minute,day,month,year,nick,
			first,last,email,reason);
}

static void 
icqlib_callback_UserFound(icq_Link* link, unsigned long uin,
       const char *nick, const char *first, const char *last,
       const char *email, char auth)
{
	callback_helper(link,name_UserFound,"(lssssb)",
			uin,nick,first,last,email,auth);
}

static void 
icqlib_callback_SearchDone(icq_Link* link)
{
	callback_helper(link,name_SearchDone,"()");
}

static void 
icqlib_callback_UserOnline(icq_Link* link, unsigned long uin,
       unsigned long status, unsigned long ip, unsigned short port,
       unsigned long real_ip, unsigned char tcp_flag)
{
	callback_helper(link,name_UserOnline,"(lllhlb)",
			uin,status,ip,port,real_ip,tcp_flag);
}

static void 
icqlib_callback_UserOffline(icq_Link* link, unsigned long uin)
{
	callback_helper(link,name_UserOffline,"(l)",
			uin);
}

static void 
icqlib_callback_UserStatusUpdate(icq_Link* link, unsigned long uin,
       unsigned long status)
{
	callback_helper(link,name_UserStatusUpdate,"(ll)",
			uin,status);
}

static void 
icqlib_callback_InfoReply(icq_Link* link, unsigned long uin,
       const char *nick, const char *first, const char *last,
       const char *email, char auth)
{
	callback_helper(link,name_InfoReply,"(lssssb)",
			uin,nick,first,last,email,auth);
}

static void 
icqlib_callback_ExtInfoReply(icq_Link* link, unsigned long uin,
       const char *city, unsigned short country_code, char country_stat,
       const char *state, unsigned short age, char gender,
       const char *phone, const char *hp, const char *about)
{
	callback_helper(link,name_ExtInfoReply,"(lshbshbsss)",
			uin,city,country_code,country_stat,
			state,age,gender,phone,hp,about);
}

static void 
icqlib_callback_WrongPassword(icq_Link* link)
{
	callback_helper(link,name_WrongPassword,"()");
}

static void 
icqlib_callback_InvalidUIN(icq_Link* link)
{
	callback_helper(link,name_InvalidUIN,"()");
}

static void 
icqlib_callback_Log(icq_Link* link, time_t time, unsigned char level,
       const char *str)
{
	callback_helper(link,name_Log,"(lcs)",
			time,level,str);
}

static void 
icqlib_callback_SrvAck(icq_Link* link, unsigned short seq)
{
	callback_helper(link,name_SrvAck,"(l)",
			seq);
}

static void 
icqlib_callback_RequestNotify(icq_Link* link, unsigned long id, int type,
			      int arg, void *data)
{
	PyObject* pyarg;

	switch (type) {
	case ICQ_NOTIFY_SUCCESS:
	case ICQ_NOTIFY_FAILED:
	case ICQ_NOTIFY_CONNECTING:
	case ICQ_NOTIFY_CONNECTED:
	case ICQ_NOTIFY_SENT:
		pyarg = PyString_FromStringAndSize(data, arg);
		break;
	case ICQ_NOTIFY_ACK:
		if (data) {
			pyarg = Py_BuildValue("(is)", arg, data);
		} else {
			pyarg = Py_BuildValue("(is)", 0, "");
		}
		break;
	case ICQ_NOTIFY_CHATSESSION:
		pyarg = PyICQChatSession_New(data);
		break;
	case ICQ_NOTIFY_FILESESSION:
		pyarg = PyICQFileSession_New(data);
		break;
	default:
		PyErr_Format(PyExc_ValueError,
			     "unknown notfiy code %d",
			     type);
		return;
	}

	callback_helper(link,name_RequestNotify,"(liO)",
			id,type,pyarg);
}

static void 
icqlib_callback_FileNotify(icq_FileSession *session, int type, int arg,
			   void *data)
{
	icq_Link* link = session->icqlink;
	PyObject* pyarg = NULL;

	switch (type) {
	case FILE_NOTIFY_DATAPACKET:
		pyarg = PyString_FromStringAndSize(data, arg);
		break;

	case FILE_NOTIFY_STATUS:
		pyarg = PyInt_FromLong(arg);
		break;

	case FILE_NOTIFY_CLOSE: {
		PyObject* dict;
		PyObject* addr;
		PyICQFileSession* pysess;
		
		dict = ((PyICQLink*)link->icq_UserData)->file_sessions;
		addr = PyInt_FromLong((long)session);
		pysess = (PyICQFileSession*)
			PyInt_AsLong(PyDict_GetItem(dict,addr));
		Py_DECREF(addr);

		pysess->fs = NULL;

		pyarg = (PyObject*)pysess;
		break;
	} 

	case FILE_NOTIFY_NEW_SPEED:
		pyarg = PyInt_FromLong(arg);
		break;

	case FILE_NOTIFY_STOP_FILE:
		pyarg = Py_None;
		Py_INCREF(Py_None);
		break;

	}

	if (type == FILE_NOTIFY_CLOSE) {
		/* invalidate corresponding PyICQFileSession */
		PyObject* dict;
		PyObject* addr;
		PyICQFileSession* pysess;
		
		dict = ((PyICQLink*)link->icq_UserData)->file_sessions;
		addr = PyInt_FromLong((long)session);
		pysess = (PyICQFileSession*)
			PyInt_AsLong(PyDict_GetItem(dict,addr));
		Py_DECREF(addr);

		pysess->fs = NULL;
	}

	callback_helper(link,name_FileNotify,"(iO)", type, pyarg);
}

static void 
icqlib_callback_ChatNotify(icq_ChatSession *session, int type, int arg,
			   void *data)
{
	icq_Link* link = session->icqlink;
	PyObject* pyarg = NULL;

	switch (type) {
	case CHAT_NOTIFY_STATUS:
		pyarg = PyInt_FromLong(arg);
		break;

	case CHAT_NOTIFY_DATA:
		pyarg = PyString_FromStringAndSize(data, arg);
		break;

	case CHAT_NOTIFY_CLOSE: {
		/* invalidate corresponding PyICQChatSession */
		PyObject* dict;
		PyObject* addr;
		PyICQChatSession* pysess;
		
		dict = ((PyICQLink*)link->icq_UserData)->chat_sessions;
		addr = PyInt_FromLong((long)session);
		pysess = (PyICQChatSession*)
			PyInt_AsLong(PyDict_GetItem(dict,addr));
		Py_DECREF(addr);

		pysess->cs = NULL;
		pyarg = (PyObject*)pysess;
		break;
	}
	}

	callback_helper(link,name_ChatNotify, "(iO)", type, pyarg);
}

static void 
icqlib_callback_NewUIN(icq_Link* link, unsigned long uin)
{
	callback_helper(link,name_NewUIN,"(l)",
			uin);
}

static void 
icqlib_callback_MetaUserFound(icq_Link* link, unsigned short seq2,
       unsigned long uin, const char *nick, const char *first,
       const char *last, const char *email, char auth)
{
	callback_helper(link,name_MetaUserFound,"(hlssssb)",
			seq2,uin,nick,first,last,email,auth);
}

static void 
icqlib_callback_MetaUserInfo(icq_Link* link, unsigned short seq2,
       const char *nick, const char *first, const char *last,
       const char *pri_eml, const char *sec_eml, const char *old_eml,
       const char *city, const char *state, const char *phone, const char *fax,
       const char *street, const char *cellular, unsigned long zip,
       unsigned short country, unsigned char timezone, unsigned char auth,
       unsigned char webaware, unsigned char hideip)
{
	/* ffs!! */
	callback_helper(link,name_MetaUserInfo,"(hsssssssssssslhbbbb)",
			seq2,nick,first,last,pri_eml,sec_eml,old_eml,
			city,state,phone,fax,street,cellular,zip,
			country,timezone,auth,webaware,hideip);
}

static void 
icqlib_callback_MetaUserWork(icq_Link* link, unsigned short seq2,
       const char *wcity, const char *wstate, const char *wphone,
       const char *wfax, const char *waddress, unsigned long wzip,
       unsigned short wcountry, const char *company, const char *department,
       const char *job, unsigned short occupation, const char *whomepage)
{
	callback_helper(link,name_MetaUserWork,"(hssssslhssshssshs)",
			seq2, wcity, wstate, wphone,wfax, waddress, wzip,
			wcountry, company, department,job, occupation, 
			whomepage);
}

static void 
icqlib_callback_MetaUserMore(icq_Link* link, unsigned short seq2,
       unsigned short age, unsigned char gender, const char *homepage,
       unsigned char byear, unsigned char bmonth, unsigned char bday,
       unsigned char lang1, unsigned char lang2, unsigned char lang3)
{
	callback_helper(link,name_MetaUserMore,"(hhbsbbbbbb)",
			seq2,age,gender,homepage,
			byear,bmonth,bday,lang1,lang2,lang3);
}

static void 
icqlib_callback_MetaUserAbout(icq_Link* link, unsigned short seq2,
       const char *about)
{
	callback_helper(link,name_MetaUserAbout,"(hs)",
			seq2,about);
}

static void 
icqlib_callback_MetaUserInterests(icq_Link* link, unsigned short seq2,
       unsigned char num, unsigned short icat1, const char *int1,
       unsigned short icat2, const char *int2, unsigned short icat3,
       const char *int3, unsigned short icat4, const char *int4)
{
	callback_helper(link,name_MetaUserInterests,"(hbhshshshs)", 
			seq2, num, icat1, int1,
			icat2, int2, icat3, int3, icat4, int4);
}

static void 
icqlib_callback_MetaUserAffiliations(icq_Link* link, unsigned short seq2,
       unsigned char anum, unsigned short acat1, const char *aff1,
       unsigned short acat2, const char *aff2, unsigned short acat3,
       const char *aff3, unsigned short acat4, const char *aff4,
       unsigned char bnum, unsigned short bcat1, const char *back1,
       unsigned short bcat2, const char *back2, unsigned short bcat3,
       const char *back3, unsigned short bcat4, const char *back4)
{
	callback_helper(link,name_MetaUserAffiliations,
			"(hbhshshshsbhshshshshs)",
			seq2, anum, acat1, aff1, acat2, aff2, acat3,
			aff3, acat4, aff4, bnum, bcat1, back1,
			bcat2, back2, bcat3, back3, bcat4, back4);
}

static void 
icqlib_callback_MetaUserHomePageCategory(icq_Link* link,
       unsigned short seq2, unsigned char num, unsigned short hcat1,
       const char *htext1)
{
	callback_helper(link,name_MetaUserHomePageCategory,"(hbhs)",
			seq2,num,hcat1,htext1);
}



/***********************************************************
 *                 ICQLink type methods                    *
 ***********************************************************/

static void
icqlib_DelPyICQLink(PyICQLink* self)
{
	if (self->link)
		icq_LinkDelete(self->link);
	PyObject_DEL(self);
}

static PyObject*
icqlib_PyICQLink_getattro(PyObject* self, PyObject* attr)
{
	PyObject* v;

	v = PyDict_GetItem(name_to_i, attr);

	if (!v) {
		return Py_FindAttr(self,attr);
	} else {
		PyObject* w = NULL;
		icq_Link* link = (icq_Link*)((PyICQLink*)self)->link;
		switch (PyInt_AS_LONG(v)) {
		case i_Uin:
			w = PyInt_FromLong(link->icq_Uin);
			break;
		case i_OurIP:
			w = PyInt_FromLong(link->icq_OurIP);
			break;
		case i_OurPort:
			w = PyInt_FromLong(link->icq_OurPort);
			break;
/*		case i_ContactList:
			w = PyInt_FromLong(link->icq_Uin);
			break;*/
		case i_Status:
			w = PyInt_FromLong(link->icq_Status);
			break;
		case i_Password:
			w = PyString_FromString(link->icq_Password);
			break;
		case i_Nick:
			w = PyString_FromString(link->icq_Nick);
			break;
		}
		return w ;
	}
}

/*
static PyObject*
icqlib_PyICQLink_setattro(PyObject* self, PyObject* attr)
{
	
}
*/

/***********************************************************
 *                 ICQLink class methods                   *
 ***********************************************************/

#define ICQLINK_DOC(FUNC) \
  static char icqlib_PyICQLink_ ## FUNC ## __doc__[]

static PyObject* 
icqlib_PyICQLink_SetProxy(PyICQLink* self, PyObject* args)
{
	char* phost;
	unsigned short pport;
	char* pname = NULL;
	char* ppass = NULL;

	if (PyTuple_Size(args) != 2 && PyTuple_Size(args) != 4) {
		PyErr_Format(PyExc_TypeError,
			     "SetProxy: expected 2 or 4 arguments, got %d",
			     PyTuple_Size(args));
		return NULL;
	}

	if (!PyArg_ParseTuple(args,"sh|ss:SetProxy",
			      &phost,&pport,&pname,&ppass)) {
		return NULL;
	}

	icq_SetProxy(self->link,phost,pport,pname != NULL,pname,ppass);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(SetProxy) = 
"SetProxy(host,port[,username,passwd]) -> None

Use |host|:|port| as your proxy, with supplied authentication information as
necessary.";

static PyObject* 
icqlib_PyICQLink_UnsetProxy(PyICQLink* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":UnsetProxy")) {
		return NULL;
	}

	icq_UnsetProxy(self->link);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(UnsetProxy) = 
"UnsetProxy() -> None

Do not use a proxy for this connection.";

static PyObject*
icqlib_PyICQLink___init__(PyICQLink* self, PyObject* args)
{
	long uin;
	char* passwd;
	char* nick;
	char use_tcp = 1;

	if (!PyArg_ParseTuple(args,"iss|O&:__init__",
			      &uin,&passwd,&nick,boolconv,&use_tcp)) {
		return NULL;
	}

	self->link = icq_LinkNew(uin,passwd,nick,use_tcp);

	if (!self->link) {
		Py_DECREF(self);
		PyErr_SetString(PyExc_MemoryError,
				"couldn't allocate icq_Link");
		return NULL;
	}

	self->chat_sessions = PyDict_New();
	if (!self->chat_sessions) {
		icq_ICQLINKDelete(self->link);
		return NULL;
	}

	self->file_sessions = PyDict_New();
	if (!self->file_sessions) {
		Py_DECREF(self->chat_sessions);
		icq_ICQLINKDelete(self->link);
		return NULL;
	}

	self->link->icq_UserData = (void*)self;
	
	self->link->icq_Logged = icqlib_callback_Logged;
	self->link->icq_Disconnected = icqlib_callback_Disconnected;
	self->link->icq_RecvMessage = icqlib_callback_RecvMessage;
	self->link->icq_RecvURL = icqlib_callback_RecvURL;
	self->link->icq_RecvWebPager = icqlib_callback_RecvWebPager;
	self->link->icq_RecvMailExpress = icqlib_callback_RecvMailExpress;
	self->link->icq_RecvChatReq = icqlib_callback_RecvChatReq;
	self->link->icq_RecvFileReq = icqlib_callback_RecvFileReq;
	self->link->icq_RecvAdded = icqlib_callback_RecvAdded;
	self->link->icq_RecvAuthReq = icqlib_callback_RecvAuthReq;
	self->link->icq_UserFound = icqlib_callback_UserFound;
	self->link->icq_SearchDone = icqlib_callback_SearchDone;
	self->link->icq_UserOnline = icqlib_callback_UserOnline;
	self->link->icq_UserOffline = icqlib_callback_UserOffline;
	self->link->icq_UserStatusUpdate = \
		icqlib_callback_UserStatusUpdate;
	self->link->icq_InfoReply = icqlib_callback_InfoReply;
	self->link->icq_ExtInfoReply = icqlib_callback_ExtInfoReply;
	self->link->icq_WrongPassword = icqlib_callback_WrongPassword;
	self->link->icq_InvalidUIN = icqlib_callback_InvalidUIN;
	self->link->icq_Log = icqlib_callback_Log;
	self->link->icq_SrvAck = icqlib_callback_SrvAck;
	self->link->icq_RequestNotify = icqlib_callback_RequestNotify;
	self->link->icq_ChatNotify = icqlib_callback_ChatNotify;
	self->link->icq_FileNotify = icqlib_callback_FileNotify;
	self->link->icq_NewUIN = icqlib_callback_NewUIN;
	self->link->icq_MetaUserFound = icqlib_callback_MetaUserFound;
	self->link->icq_MetaUserInfo = icqlib_callback_MetaUserInfo;
	self->link->icq_MetaUserWork = icqlib_callback_MetaUserWork;
	self->link->icq_MetaUserMore = icqlib_callback_MetaUserMore;
	self->link->icq_MetaUserAbout = icqlib_callback_MetaUserAbout;
	self->link->icq_MetaUserInterests = \
		icqlib_callback_MetaUserInterests;
	self->link->icq_MetaUserAffiliations = \
		icqlib_callback_MetaUserAffiliations;
	self->link->icq_MetaUserHomePageCategory = \
		icqlib_callback_MetaUserHomePageCategory;

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(__init__) = "\
ICQLink(uin, passwd, nick) -> new ICQLink object\n\
\n\
Initialise a new ICQLink with the obviously named parameters.";

static PyObject* 
icqlib_PyICQLink_Connect(PyICQLink* self, PyObject* args)
{
	char* host = "icq.mirabilis.com";
	int port = 4000;

	if (!PyArg_ParseTuple(args,"|si:Connect",&host,&port)) {
		return NULL;
	}

	return PyInt_FromLong(icq_Connect(self->link,host,port));
}

ICQLINK_DOC(Connect) = 
"Connect(host=\"icq.mirabilis.com\", port=4000) -> int\n\
\n\
Connect to the server at |host|:|port|.  Return the fileno of the\n\
connecting socket (which can also be retreived with GetSok, so\n\
there's no need to be too carfeul about squirreling it away).";

static PyObject*
icqlib_PyICQLink_Disconnect(PyICQLink* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":Disconnect")) {
		return NULL;
	}

	icq_Disconnect(self->link);
	
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(Disconnect) = 
"Disconnect() -> None\n\
\n\
Disconnect from the server";

static PyObject* 
icqlib_PyICQLink_GetSok(PyICQLink* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":GetSok")) {
		return NULL;
	}

	return PyInt_FromLong(icq_GetSok(self->link));
}

ICQLINK_DOC(GetSok) =
"GetSok() -> int\n\
\n\
return the file descriptor of the socket connecting us to the\n\
server.";

static PyObject* 
icqlib_PyICQLink_GetProxySok(PyICQLink* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":GetProxySok")) {
		return NULL;
	}

	return PyInt_FromLong(icq_GetProxySok(self->link));
}

ICQLINK_DOC(GetProxySok) =
"GetProxySok() -> int\n\
\n\
return the file descriptor of the socket connecting us to the\n\
proxy server.";

static PyObject* 
icqlib_PyICQLink_HandleServerResponse(PyICQLink* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":HandleServerResponse")) {
		return NULL;
	}

	icq_HandleServerResponse(self->link);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(HandleServerResponse) =
"HandleServerResponse() -> None\n\
\n\
??";

static PyObject* 
icqlib_PyICQLink_HandleProxyResponse(PyICQLink* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":HandleProxyResponse")) {
		return NULL;
	}

	icq_HandleProxyResponse(self->link);

	if (PyErr_Occurred()) {
		return NULL;
	} else {
		Py_INCREF(Py_None);
		return Py_None;
	}	
}

ICQLINK_DOC(HandleProxyResponse) =
"HandleProxyResponse() -> None\n\
\n\
??";

static PyObject* 
icqlib_PyICQLink_KeepAlive(PyICQLink* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":KeepAlive")) {
		return NULL;
	}

	return PyInt_FromLong(icq_KeepAlive(self->link));
}

ICQLINK_DOC(KeepAlive) =
"KeepAlive() -> int\n\
\n\
Tell the server we're still here - must be called every 2 minutes or\n\
so.  Returns a sequence number whose significance is less than clear\n\
to me.";

static PyObject*
icqlib_PyICQLink_Login(PyICQLink* self, PyObject* args)
{
	long status = STATUS_ONLINE;

	if (!PyArg_ParseTuple(args,"|l:Login",&status)) {
		return NULL;
	}

	icq_Login(self->link,status);
	
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(Login) =
"Login(status=STATUS_ONLINE) -> None\n\
\n\
Attempt to login to the server.  Must be called after Connect.  This\n\
returns immediately - you will learn the success or failure or the\n\
login from callbacks.";

static PyObject*
icqlib_PyICQLink_Logout(PyICQLink* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":Logout")) {
		return NULL;
	}

	icq_Logout(self->link);
	
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(Logout) =
"Logout() -> None\n\
\n\
What do you think this function does?";

static PyObject* 
icqlib_PyICQLink_SendContactList(PyICQLink* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":SendContactList")) {
		return NULL;
	}

	icq_SendContactList(self->link);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(SendContactList) =
"SendContactList() -> None\n\
\n\
Syncronise the server's idea of the contact list with the client's.\n\
You will get UserOnline/UserStatusUpdate/UserOffline callbacks for\n\
users on your contact list, which can be manipulated with ContactAdd\n\
and ContactRemove.";

static PyObject* 
icqlib_PyICQLink_SendVisibleList(PyICQLink* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":SendVisibleList")) {
		return NULL;
	}

	icq_SendVisibleList(self->link);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(SendVisibleList) =
"SendVisibleList() -> None\n\
\n\
Syncronise the server's idea of the visible list with the client's.\n\
Users on your visible list can see you when you are invisble.  You\n\
can manipulate the visible list with ContactSetVis.";

static PyObject*
icqlib_PyICQLink_SendInvisibleList(PyICQLink* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":SendInvisibleList")) {
		return NULL;
	}

	icq_SendInvisibleList(self->link);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(SendInvisibleList) =
"SendInvisibleList() -> None\n\
\n\
Syncronise the server's idea of the invisible list with the client's.\n\
Users on your invisible list cannot see you, even when you are visble.\n\
You can manipulate the invisible list with ContactSetInvis.";

static PyObject* 
icqlib_PyICQLink_SendNewUser(PyICQLink* self, PyObject* args)
{
	unsigned long uin;

	if (!PyArg_ParseTuple(args,"l:SendNewUser",&uin)) {
		return NULL;
	}

	icq_SendNewUser(self->link,uin);


	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(SendNewUser) =
"SendNewUser() -> None\n\
\n\
not sure how this works, sorry.";

static PyObject* 
icqlib_PyICQLink_SendMessage(PyICQLink* self, PyObject* args)
{
	long uin;
	char* message;
	int method = ICQ_SEND_BESTWAY;

	if (!PyArg_ParseTuple(args,"ls|i:SendMessage",
			      &uin,&message,&method)) {
		return NULL;
	}

	return PyInt_FromLong(icq_SendMessage(self->link,
					      uin,message,method));
}

ICQLINK_DOC(SendMessage) = 
"SendMessage(uin, message[, method]) -> int\n\
\n\
This is icq!  You know what this function does!  |method| can be\n\
one of ICQLIB.ICQ_SEND_THRUSERVER, ICQLIB.ICQ_SEND_DIRECT or\n\
ICQLIB.ICQ_SEND_BESTWAY (which is the default).  Returns a sequence\n\
number which can be matched up with a callback to RequestNotify.";

static PyObject* 
icqlib_PyICQLink_SendURL(PyICQLink* self, PyObject* args)
{
	long uin;
	char* url;
	char* descr;
	int thru_server = ICQ_SEND_BESTWAY;

	if (!PyArg_ParseTuple(args,"lss|i:SendURL",
			      &uin,&url,&descr,&thru_server)) {
		return NULL;
	}

	return PyInt_FromLong(icq_SendURL(self->link,
					  uin,url,descr,thru_server));
}

ICQLINK_DOC(SendURL) = 
"SendURL(uin, url, descr[, method]) -> int
\n
This is icq!  You know what this function does!  |method| can be one\n\
of ICQLIB.ICQ_SEND_THRUSERVER, ICQLIB.ICQ_SEND_DIRECT or\n\
ICQLIB.ICQ_SEND_BESTWAY (which is the default).  Returns a sequence\n\
number which can be matched up with a callback to RequestNotify.";

static PyObject*
icqlib_PyICQLink_ChangeStatus(PyICQLink* self, PyObject* args)
{
	long status;

	if (!PyArg_ParseTuple(args,"l:ChangeStatus",&status)) {
		return NULL;
	}

	icq_ChangeStatus(self->link,status);
	
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(ChangeStatus) =
"ChangeStatus(status) -> None

Change you status to, well, |status|.  I don't really understand how the
values in ICQLIB match up with what you want to pass in here, but I'm sure you
can work it out ;-).";

static PyObject* 
icqlib_PyICQLink_SendInfoReq(PyICQLink* self, PyObject* args)
{
	unsigned long uin;

	if (!PyArg_ParseTuple(args,"l:SendInfoReq",&uin)) {
		return NULL;
	}

	return PyInt_FromLong(icq_SendInfoReq(self->link,uin));
}

ICQLINK_DOC(SendInfoReq) = 
"SendInfoReq(uin) -> int

Request info about user |uin|; InfoReply will be called in due course.  Not
sure what the return value's for...";

static PyObject* 
icqlib_PyICQLink_SendExtInfoReq(PyICQLink* self, PyObject* args)
{
	unsigned long uin;

	if (!PyArg_ParseTuple(args,"l:SendExtInfoReq",&uin)) {
		return NULL;
	}

	return PyInt_FromLong(icq_SendExtInfoReq(self->link,uin));
}

ICQLINK_DOC(SendExtInfoReq) = 
"SendExtInfoReq(uin) -> int

Request more info about user |uin|; ExtInfoReply will be called in due course.
Not sure what the return value's for...";

static PyObject* 
icqlib_PyICQLink_SendAuthMsg(PyICQLink* self, PyObject* args)
{
	unsigned long uin;

	if (!PyArg_ParseTuple(args,"l:SendAuthMsg",&uin)) {
		return NULL;
	}

	return PyInt_FromLong(icq_SendAuthMsg(self->link,uin));
}

ICQLINK_DOC(SendAuthMsg) =
"SendAuthMsg(uin) -> int\n\
\n\
Allow user |uin| to add you to they're contact list.  Call in response\n\
to a call to RecvAuthReq.\n\
\n\
(This is one of the more amusing \"security\" features of icq - the\n\
official Mirabilis client is generally the only one to actually\n\
ask...).  Returns a sequence number.";

static PyObject* 
icqlib_PyICQLink_SendSearchReq(PyICQLink* self, PyObject* args)
{
	char* email;
	char* nick;
	char* first;
	char* last;

	if (!PyArg_ParseTuple(args,"zzzz:SendSearchReq",
			      &email,&nick,&first,&last)) {
		return NULL;
	}

	icq_SendSearchReq(self->link,email,nick,first,last);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(SendSearchReq) =
"SendSearchReq(email,nick,first,last) -> None

Search for a user with specified details (which may be None).
UserFound will be called for each matching user, then SearchDone.";

static PyObject* 
icqlib_PyICQLink_SendSearchUINReq(PyICQLink* self, PyObject* args)
{
	unsigned long uin;

	if (!PyArg_ParseTuple(args,"l:SendSearchUINReq",&uin)) {
		return NULL;
	}

	icq_SendSearchUINReq(self->link,uin);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(SendSearchUINReq) = 
"SendSearchUINReq(uin) -> None

Search for a user by uin.  I imagine this also results in calls to
UserFound and SearchDone.";

static PyObject* 
icqlib_PyICQLink_RegNewUser(PyICQLink* self, PyObject* args)
{
	char *pass;

	if (!PyArg_ParseTuple(args,"l:RegNewUser",&pass)) {
		return NULL;
	}

	icq_RegNewUser(self->link,pass);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(RegNewUser) =
"RegNewUser(passwd) -> None

Not sure how this works.";

static PyObject* 
icqlib_PyICQLink_UpdateUserInfo(PyICQLink* self, PyObject* args)
{
	char* nick;
	char* first;
	char* last;
	char* email;

	if (!PyArg_ParseTuple(args,"ssss:UpdateUserInfo",
			      &nick,&first,&last,&email)) {
		return NULL;
	}

	return PyInt_FromLong(icq_UpdateUserInfo(self->link,
						 nick,first,last,email));
}

ICQLINK_DOC(UpdateUserInfo) =
"UpdateUserInfo(nick,first,last,email) -> sequence number

Update your information in the server's user database.";

static PyObject*
icqlib_PyICQLink_UpdateAuthInfo(PyICQLink* self, PyObject* args)
{
	char auth;

	if (!PyArg_ParseTuple(args,"O&:UpdateAuthInfo",
			      boolconv,&auth)) {
		return NULL;
	}


	return PyInt_FromLong(icq_UpdateAuthInfo(self->link,auth));
}

ICQLINK_DOC(UpdateAuthInfo) =
"UpdateAuthInfo(auth) -> sequence number

Update your authorization status. (i.e. do you want users to require your
permission to add you to their contact lists?) Note that this only concerns
the Mirabilis client, most clones allow their users to override this setting.";

static PyObject*
icqlib_PyICQLink_UpdateMetaInfoSet(PyICQLink* self, PyObject* args)
{
	char *nick;
	char *first;
	char *last;
	char *email;
	char *email2;
	char *email3;
	char *city;
	char *state;
	char *phone;
	char *fax;
	char *street;
	char *cellular;
	unsigned long zip;
	unsigned short cnt_code;
	unsigned char cnt_stat;
	unsigned char emailhide;

	if (!PyArg_ParseTuple(args,"sssssssssssslhbb:UpdateMetaInfoSet",
			      &nick, &first, &last, &email,
			      &email2, &email3, &city,
			      &state, &phone, &fax, &street,
			      &cellular, &zip, &cnt_code,
			      &cnt_stat,  &emailhide)) {
		return NULL;
	}

	return PyInt_FromLong(icq_UpdateMetaInfoSet(self->link,
						    nick, first, last, email,
						    email2, email3, city,
						    state, phone, fax, street,
						    cellular, zip, cnt_code,
						    cnt_stat,  emailhide));
}

ICQLINK_DOC(UpdateMetaInfoSet) =
"UpdateMetaInfoSet(nick,first,last,email,email2,email3,city,state,phone,fax,
                  street,cellular,zip,cnt_code,cnt_stat,emailhide) -> int

Lose your privacy!  Returns a sequence number.";

static PyObject*
icqlib_PyICQLink_UpdateMetaInfoHomepage(PyICQLink* self, PyObject* args)
{
	unsigned char age;
	char *homepage;
	unsigned char year;
	unsigned char month;
	unsigned char day;
	unsigned char lang1;
	unsigned char lang2;
	unsigned char lang3;

	if (!PyArg_ParseTuple(args,"bsbbbbbb:UpdateMetaInfoHomepage",
			      &age,&homepage,&year,&month,&day,
			      &lang1,&lang2,&lang3)) {
		return NULL;
	}

	return PyInt_FromLong(icq_UpdateMetaInfoHomepage(
		self->link,
		age,homepage,year,month,day,
		lang1,lang2,lang3));
}

ICQLINK_DOC(UpdateMetaInfoHomepage) =
"UpdateMetaInfoHomepage(age,homepage,year,month,day,
                       lang1,lang2,lang3) -> sequence number

Lose your privacy!  Returns a sequence number.";

static PyObject*
icqlib_PyICQLink_UpdateMetaInfoSecurity(PyICQLink* self, PyObject* args)
{
	char reqauth,webpresence,pubip;

	if (!PyArg_ParseTuple(args,"bbb:UpdateMetaInfoSecurity",
			      &reqauth,&webpresence,&pubip)) {
		return NULL;
	}

	return PyInt_FromLong(icq_UpdateMetaInfoSecurity(
		self->link,reqauth,webpresence,pubip));
}

ICQLINK_DOC(UpdateMetaInfoSecurity) =
"UpdateMetaInfoSecurity(reqauth,webpresence,pubip) -> int

Retain your privacy!  But only with users of official clients!  Returns a
sequence number.";

static PyObject*
icqlib_PyICQLink_UpdateMetaInfoAbout(PyICQLink* self, PyObject* args)
{
	char* about;

	if (!PyArg_ParseTuple(args,"s:UpdateMetaInfoAbout",
			      &about)) {
		return NULL;
	}

	return PyInt_FromLong(icq_UpdateMetaInfoAbout(
		self->link,about));
}

ICQLINK_DOC(UpdateMetaInfoAbout) =
"UpdateMetaInfoAbout(about) -> int

Lose you privacy!  Returns a sequence number.";

static PyObject*
icqlib_PyICQLink_UpdateNewUserInfo(PyICQLink* self, PyObject* args)
{
	char* nick;
	char* first;
	char* last;
	char* email;

	if (!PyArg_ParseTuple(args,"sss:UpdateNewUserInfo",
			      &nick,&first,&last,&email)) {
		return NULL;
	}

	return PyInt_FromLong(icq_UpdateNewUserInfo(
		self->link,nick,first,last,email));
}

ICQLINK_DOC(UpdateNewUserInfo) =
"UpdateNewUserInfo(nick,first,last,email) -> int

Not sure what this does, except return a sequence number.";

static PyObject*
icqlib_PyICQLink_SendMetaInfoReq(PyICQLink* self, PyObject* args)
{
	unsigned long uin;

	if (!PyArg_ParseTuple(args,"l:SendMetaInfoReq",
			      &uin)) {
		return NULL;
	}

	return PyInt_FromLong(icq_SendMetaInfoReq(
		self->link,uin));
}

ICQLINK_DOC(SendMetaInfoReq) =
"SendMetaInfoReq(uin) -> int

Arrange for callbacks to MetaUser*.  Returns a sequnce number.";

/*FmtLog*/

static PyObject* 
icqlib_PyICQLink_ContactAdd(PyICQLink* self, PyObject* args)
{
	unsigned long cuin;

	if (!PyArg_ParseTuple(args,"l:ContactAdd",&cuin)) {
		return NULL;
	}

	icq_ContactAdd(self->link,cuin);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(ContactAdd) =
"ContactAdd(cuin) -> None

Add user |cuin| to your contact list; this does not affect the server's idea
of your contact list until you call SendContactList.";

static PyObject* 
icqlib_PyICQLink_ContactRemove(PyICQLink* self, PyObject* args)
{
	unsigned long cuin;

	if (!PyArg_ParseTuple(args,"l:ContactRemove",&cuin)) {
		return NULL;
	}

	icq_ContactRemove(self->link,cuin);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(ContactRemove) =
"ContactRemove(cuin) -> None

Remove user |cuin| from your contact list; this does not affect the server's
idea of your contact list until you call SendContactList.";

static PyObject* 
icqlib_PyICQLink_ContactClear(PyICQLink* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":ContactClear")) {
		return NULL;
	}

	icq_ContactClear(self->link);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(ContactClear) =
"ContactClear(cuin) -> None

Remove all users from your contact list; this does not affect the server's
idea of your contact list until you call SendContactList.";

static PyObject* 
icqlib_PyICQLink_ContactSetVis(PyICQLink* self, PyObject* args)
{
	unsigned long cuin;
	char vu;

	if (!PyArg_ParseTuple(args,"lO&:ContactSetVis",&cuin,boolconv,&vu)) {
		return NULL;
	}

	icq_ContactSetVis(self->link,cuin,vu);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(ContactSetVis) =
"ContactSetVis(cuin,vis) -> None

Add or remove user |cuin| from your visible list as determined by |vis|; this
does not affect the server's idea of your visible list until you call
SendVisibleList.  A user on your visible list can see you when you are
invisible.  Has no effect at all if |cuin| is not on your contact list.";

static PyObject* 
icqlib_PyICQLink_ContactSetInvis(PyICQLink* self, PyObject* args)
{
	unsigned long cuin;
	char vu;

	if (!PyArg_ParseTuple(args,"lO&:ContactSetInvis",&cuin,boolconv,&vu)) {
		return NULL;
	}

	icq_ContactSetInvis(self->link,cuin,vu);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(ContactSetInvis) =
"ContactSetInvis(cuin,invis) -> None

Add or remove user |cuin| from your visible list as determined by |invis|;
this does not affect the server's idea of your visible list until you call
SendInvisibleList.  A user on your invisible list cannot see you when you are
visible.  Has no effect at all if |cuin| is not on your contact list.";

/*
static PyObject* 
icqlib_PyICQLink_TCPProcessReceived(PyICQLink* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":TCPProcesReceived")) {
		return NULL;
	}

	icq_TCPProcessReceived(self->link);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(TCPProcessReceived) =
"TCPProcessReceived() -> None

This function doesn't seem to exist.  Confused?  So am I.";
*/

static PyObject* 
icqlib_PyICQLink_TCPSendMessage(PyICQLink* self, PyObject* args)
{
	long uin;
	char* message;

	if (!PyArg_ParseTuple(args,"ls:TCPSendMessage",
			      &uin,&message)) {
		return NULL;
	}

	return PyInt_FromLong(icq_TCPSendMessage(
		self->link,uin,message));
}

ICQLINK_DOC(TCPSendMessage) =
"TCPSendMessage(uin,message) -> sequence number

Send a message via TCP (as opposed to, presumably, UDP).  Don't know why you'd
want to do this, but isn't it nice that you can?";

static PyObject* 
icqlib_PyICQLink_TCPSendURL(PyICQLink* self, PyObject* args)
{
	long uin;
	char* url;
	char* descr;

	if (!PyArg_ParseTuple(args,"lss:TCPSendURL",
			      &uin,&url,&descr)) {
		return NULL;
	}

	return PyInt_FromLong(icq_TCPSendURL(
		self->link,uin,url,descr));
}

ICQLINK_DOC(TCPSendURL) =
"TCPSendURL(uin,url,descr) -> sequence number

Send a URL via TCP (as opposed to, presumably, UDP).  Don't know why you'd
want to do this, but isn't it nice that you can?";
static PyObject*
icqlib_PyICQLink_AcceptFileRequest(PyICQLink* self, PyObject* args)
{
	unsigned long cuin, seq;
	icq_FileSession* fs;

	if (!PyArg_ParseTuple(args,"ll:AcceptFileRequest",&cuin,&seq)) {
		return NULL;
	}
	
	fs = icq_AcceptFileRequest(self->link,cuin,seq);

	if (!fs) {
		PyErr_SetString(PyExc_RuntimeError,
			       "icq_AcceptFileRequest returned NULL");
		return NULL;
	}

	return PyICQFileSession_New(fs); 
}

ICQLINK_DOC(AcceptFileRequest) =
"AcceptFileRequest(cuin,seq) -> ICQFileSession

OK: dodgy terminology - in accepting a file request, you are agreeing to
*receive* some files!  I'm not going to document file transfers yet as I
haven't even tried to use them.";

static PyObject*
icqlib_PyICQLink_SendFileRequest(PyICQLink* self, PyObject* args)
{
	unsigned long uin;
	char* message;
	char** array;
	PyObject* list;
	PyObject* ret;
	int i;

	if (!PyArg_ParseTuple(args,"lsO:SendFileRequest",
			      &uin,&message,&list)) {
		return NULL;
	}
	
	if (!PyList_Check(list)) {
		PyErr_Format(PyExc_TypeError,
			"SendFileRequest: arg 1 must be a list, not a %.500s",
			     list->ob_type->tp_name);
		return NULL;
	}

	array = malloc(sizeof(char*) * (PyList_GET_SIZE(list)));
	
	if (!array) {
		PyErr_SetString(PyExc_MemoryError,
				"SendFileRequest: out of memory!");
		return NULL;
	}

	for (i = 0; i < PyList_GET_SIZE(list); i++) {
		array[i] = PyString_AsString(PyList_GET_ITEM(list,i));
		if (!array[i]) {
			free(array);
			return NULL;
		}
	}

	array[PyList_GET_SIZE(list)] = NULL;

	ret = PyInt_FromLong(icq_SendFileRequest(
		self->link,uin,message,array));
	
	free(array);

	return ret;
}

ICQLINK_DOC(SendFileRequest) =
"SendFileRequest(cuin,msg,files) -> int

OK: dodgy terminology - a file request is a request to *send* some files!  So
this function asks |cuin| if they would like to receive |files| - which must
be a list of filenames.  I'm not going to document file transfers yet as I
haven't even tried to use them.";

static PyObject*
icqlib_PyICQLink_CancelFileRequest(PyICQLink* self, PyObject* args)
{
	unsigned long uin, seq;

	if (!PyArg_ParseTuple(args,"ll:CancelFileRequest",&uin,&seq)) {
		return NULL;
	}

	icq_CancelFileRequest(self->link,uin,seq);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(CancelFileRequest) =
"CancelFileRequest(cuin,seq) -> None

Give up waiting for user |uin| to respond to your file request.";

static PyObject*
icqlib_PyICQLink_RefuseFileRequest(PyICQLink* self, PyObject* args)
{
	unsigned long uin, seq;
	char* reason;

	if (!PyArg_ParseTuple(args,"lls:RefuseFileRequest",
			      &uin,&seq,&reason)) {
		return NULL;
	}

	icq_RefuseFileRequest(self->link,uin,seq,reason);

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(RefuseFileRequest) =
"RefuseFileRequest(uin,seq,reason) -> None

Refuse the file request from |uin|; get |seq| from a call to RecvFileReq and
say why in |reason|.";

/***********************************************************
 *                 ICQLink callback stubs                  *
 ***********************************************************/

static PyObject*
icqlib_PyICQLink_Logged(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(Logged) =
"[callback] Logged()

This is called when the server acknowledges a successful login.";

static PyObject*
icqlib_PyICQLink_Disconnected(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(Disconnected) =
"[callback] Disconnected()

This is called when the server boots us off for some reason.";

static PyObject*
icqlib_PyICQLink_RecvMessage(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(RecvMessage) =
"[callback] RecvMessage(uin,(hour,minute,day,month,year),message)

Called when a message is recieved from |uin|.";

static PyObject*
icqlib_PyICQLink_RecvURL(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(RecvURL) =
"[callback] RecvURL(uin,(hour,minute,day,month,year),url,descr)

Called when a url is recieved from |uin|.";

static PyObject*
icqlib_PyICQLink_RecvWebPager(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(RecvWebPager) =
"[callback] RecvWebPager((hour,minute,day,month,year),nick,email,message)

Called when a web pager message is received.";

static PyObject*
icqlib_PyICQLink_RecvMailExpress(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(RecvMailExpress) =
"[callback] RecvMailExpress((hour,minute,day,month,year),nick,email,message)

I don't know what MailExpress is!";

static PyObject*
icqlib_PyICQLink_RecvChatReq(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(RecvChatReq) =
"[callback] RecvChatReq(uin,(hour,minute,day,month,year),descr,seq)

Called when user |uin| requests a chat with you - save |seq| for calls to
AcceptChatRequest or RefuseChatRequest as appropriate.";

static PyObject*
icqlib_PyICQLink_RecvFileReq(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(RecvFileReq) =
"[callback] RecvFileReq(uin,(hour,minute,day,month,year),
                       descr,filename,filesize,seq)

Called when user |uin| wants to send |filename| to you - save |seq| for calls
to AccpetFileRequest or RefuseFileRequest as appropriate.";

static PyObject*
icqlib_PyICQLink_RecvAdded(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(RecvAdded) =
"[callback] RecvAdded(uin,(hour,minute,day,month,year),nick,first,last,email)

Called when user |uin| adds you to their contact list.";

static PyObject*
icqlib_PyICQLink_RecvAuthReq(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(RecvAuthReq) =
"[callback] RecvAuthReq(uin,(hour,minute,day,month,year),nick,
                       first,last,email,reason)

User |uin| wants to add you to their contact list - will you let them?";

static PyObject*
icqlib_PyICQLink_UserFound(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(UserFound) =
"[callback] UserFound(uin,nick,first,last,email,auth)

Called in repsonse to a call to SendSearchUINReq and SendSearchReq - you will
get a call for each matching user.";

static PyObject*
icqlib_PyICQLink_SearchDone(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(SearchDone) =
"[callback] SearchDone()

Called when the server has run out of users matching the criteria to a
call to SendSearchUINReq and SendInfoReq - or it the request matched
\"too many\" users.";

static PyObject*
icqlib_PyICQLink_UserOnline(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(UserOnline) =
"[callback] UserOnline(uin,status,ip,port,real_ip,tcp_flag)

Called when a user on your contact list becomes visible to you.";

static PyObject*
icqlib_PyICQLink_UserOffline(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(UserOffline) =
"[callback] UserOffline(uin)

Called when a user on your contact list ceases to be visible to you.";

static PyObject*
icqlib_PyICQLink_UserStatusUpdate(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(UserStatusUpdate) =
"[callback] UserStatusUpdate(uin,status)

Called when a user on your contact list changes status (duh!).  The
interpretation of the |status| field is not entirely straightforward,
sadly.";

static PyObject*
icqlib_PyICQLink_InfoReply(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(InfoReply) =
"[callback] InfoReply(uin,nick,first,last,email,auth)

Called in repsonse to a call to SendInfoReq.";

static PyObject*
icqlib_PyICQLink_ExtInfoReply(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(ExtInfoReply) =
"[callback] ExtInfoReply(uin,city,country_code,country_stat,
                        state,age,gender,phone,hp,about)

Called in repsonse to a call to SendExtInfoReq.";

static PyObject*
icqlib_PyICQLink_WrongPassword(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(WrongPassword) =
"[callback] WrongPassword()

Called when the server refuses your login because of a wrong password.";

static PyObject*
icqlib_PyICQLink_InvalidUIN(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(InvalidUIN) =
"[callback] InvalidUIN()

Called when the server refuses your login because of an invalid uin.";

static PyObject*
icqlib_PyICQLink_Log(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(Log) =
"[callback] Log(time,level,str)

Called when icqlib thinks something's worth logging - cf. the
Get/SetLogLevel module level functions...";

static PyObject*
icqlib_PyICQLink_SrvAck(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(SrvAck) =
"[callback] SrvAck(seq)

The server has acknowledged you!  How nice!";

static PyObject*
icqlib_PyICQLink_RequestNotify(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(RequestNotify) =
"[callback] RequestNotify(id,result,data)

|id| is a sequence number that will have been returned by some ICQLink
method; |result| is one of the ICQLIB.ICQ_NOTIFY_* constants, and
|data| usually seems to be the empty string.";

static PyObject*
icqlib_PyICQLink_NewUIN(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(NewUIN) =
"[callback] NewUIN(uin)

I don't understand new user voodoo.";

static PyObject*
icqlib_PyICQLink_MetaUserFound(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(MetaUserFound) =
"[callback] MetaUserFound(...)

Argh!  Too many arguments!  Can't be bothered!";

static PyObject*
icqlib_PyICQLink_MetaUserInfo(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(MetaUserInfo) =
"[callback] MetaUserInfo(...)

Argh!  Too many arguments!  Can't be bothered!";

static PyObject*
icqlib_PyICQLink_MetaUserWork(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(MetaUserWork) =
"[callback] MetaUserWork(...)

Argh!  Too many arguments!  Can't be bothered!";

static PyObject*
icqlib_PyICQLink_MetaUserMore(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(MetaUserMore) =
"[callback] MetaUserMore(...)

Argh!  Too many arguments!  Can't be bothered!";

static PyObject*
icqlib_PyICQLink_MetaUserAbout(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(MetaUserAbout) =
"[callback] MetaUserAbout(...)

Argh!  Too many arguments!  Can't be bothered!";

static PyObject*
icqlib_PyICQLink_MetaUserInterests(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(MetaUserInterests) =
"[callback] MetaUserInterests(...)

Argh!  Too many arguments!  Can't be bothered!";

static PyObject*
icqlib_PyICQLink_MetaUserAffiliations(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(MetaUserAffiliations) =
"[callback] MetaUserAffiliations(...)

Argh!  Too many arguments!  Can't be bothered!";

static PyObject*
icqlib_PyICQLink_MetaUserHomePageCategory(PyObject* self, PyObject* args)
{
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLINK_DOC(MetaUserHomePageCategory) =
"[callback] MetaUserHomePageCategory(...)

Argh!  Too many arguments!  Can't be bothered!";

#define PYICQLINK_METH(NAME) \
  { #NAME, (PyCFunction)icqlib_PyICQLink_ ## NAME, METH_VARARGS, \
   icqlib_PyICQLink_ ## NAME ## __doc__ }

static PyMethodDef icqlib_PyICQLink_methods[] = {
	PYICQLINK_METH(SetProxy),
	PYICQLINK_METH(UnsetProxy),

	PYICQLINK_METH(__init__),
	PYICQLINK_METH(Connect),
	PYICQLINK_METH(Disconnect),
	PYICQLINK_METH(GetSok),
	PYICQLINK_METH(GetProxySok),
	PYICQLINK_METH(HandleServerResponse),
	PYICQLINK_METH(HandleProxyResponse),
	PYICQLINK_METH(KeepAlive),
	PYICQLINK_METH(Login),
	PYICQLINK_METH(Logout),
	PYICQLINK_METH(SendContactList),
	PYICQLINK_METH(SendVisibleList),
	PYICQLINK_METH(SendInvisibleList),
	PYICQLINK_METH(SendNewUser),
	PYICQLINK_METH(SendMessage),
	PYICQLINK_METH(SendURL),
	PYICQLINK_METH(ChangeStatus),
	PYICQLINK_METH(SendInfoReq),
	PYICQLINK_METH(SendExtInfoReq),
	PYICQLINK_METH(SendAuthMsg),
	PYICQLINK_METH(SendSearchReq),
	PYICQLINK_METH(SendSearchUINReq),

	PYICQLINK_METH(RegNewUser),
	PYICQLINK_METH(UpdateUserInfo),
	PYICQLINK_METH(UpdateAuthInfo),
	PYICQLINK_METH(UpdateMetaInfoSet),
	PYICQLINK_METH(UpdateMetaInfoHomepage),
	PYICQLINK_METH(UpdateMetaInfoAbout),
	PYICQLINK_METH(UpdateMetaInfoSecurity),
	PYICQLINK_METH(UpdateNewUserInfo),
	PYICQLINK_METH(SendMetaInfoReq),

/*	PYICQLINK_METH(FmtLog),*/

	PYICQLINK_METH(ContactAdd),
	PYICQLINK_METH(ContactRemove),
	PYICQLINK_METH(ContactClear),
	PYICQLINK_METH(ContactSetVis),
	PYICQLINK_METH(ContactSetInvis),

	PYICQLINK_METH(TCPSendMessage),
	PYICQLINK_METH(TCPSendURL),
	
	PYICQLINK_METH(AcceptFileRequest),
	PYICQLINK_METH(SendFileRequest),
	PYICQLINK_METH(CancelFileRequest),
	PYICQLINK_METH(RefuseFileRequest),

/**************callbacks**************/

	PYICQLINK_METH(Logged),
	PYICQLINK_METH(Disconnected),
	PYICQLINK_METH(RecvMessage),
	PYICQLINK_METH(RecvURL),
	PYICQLINK_METH(RecvWebPager),
	PYICQLINK_METH(RecvMailExpress),
	PYICQLINK_METH(RecvChatReq),
	PYICQLINK_METH(RecvFileReq),
	PYICQLINK_METH(RecvAdded),
	PYICQLINK_METH(RecvAuthReq),
	PYICQLINK_METH(UserFound),
	PYICQLINK_METH(SearchDone),
	PYICQLINK_METH(UserOnline),
	PYICQLINK_METH(UserOffline),
	PYICQLINK_METH(UserStatusUpdate),
	PYICQLINK_METH(InfoReply),
	PYICQLINK_METH(ExtInfoReply),
	PYICQLINK_METH(WrongPassword),
	PYICQLINK_METH(InvalidUIN),
	PYICQLINK_METH(Log),
	PYICQLINK_METH(SrvAck),
	PYICQLINK_METH(RequestNotify),
	PYICQLINK_METH(NewUIN),
	PYICQLINK_METH(MetaUserFound),
	PYICQLINK_METH(MetaUserInfo),
	PYICQLINK_METH(MetaUserWork),
	PYICQLINK_METH(MetaUserMore),
	PYICQLINK_METH(MetaUserAbout),
	PYICQLINK_METH(MetaUserInterests),
	PYICQLINK_METH(MetaUserAffiliations),
	PYICQLINK_METH(MetaUserHomePageCategory),
	
	{NULL,NULL}
};

static PyExtensionClass PyICQLinkType = {
	PyObject_HEAD_INIT(NULL)
	0,
	"PyICQLink",
        sizeof(PyICQLink),      /*tp_basicsize*/
        0,                      /*tp_itemsize*/
        /* methods */
        (destructor)icqlib_DelPyICQLink, /*tp_dealloc*/
        0,                               /*tp_print*/
        0,                               /*tp_getattr*/
        0,                               /*tp_setattr*/
        0,                      /*tp_compare*/
        0,                      /*tp_repr*/
        0,                      /*tp_as_number*/
        0,                      /*tp_as_sequence*/
        0,                      /*tp_as_mapping*/
        0,                      /*tp_hash*/
	0, /*tp_call*/
	0, /*tp_str*/
	(getattrofunc)icqlib_PyICQLink_getattro, /*tp_getattro*/
	(setattrofunc)/*icqlib_PyICQLink_setattro*/0, /*tp_setattro*/
	0, 0, /*reservered*/

	"",
	METHOD_CHAIN(icqlib_PyICQLink_methods),
};

/***********************************************************
 *                     ICQFileSession                      *
 ***********************************************************/

static PyObject* fs_name_id;
#define fs_i_id                      0
static PyObject* fs_name_status;
#define fs_i_status                  1
static PyObject* fs_name_icqlink;
#define fs_i_icqlink                 2
static PyObject* fs_name_direction;
#define fs_i_direction               4
static PyObject* fs_name_remote_uin;
#define fs_i_remote_uin              5
static PyObject* fs_name_remote_handle;
#define fs_i_remote_handle           6
static PyObject* fs_name_files;
#define fs_i_files                   7
static PyObject* fs_name_total_files;
#define fs_i_total_files             8
static PyObject* fs_name_current_file_num;
#define fs_i_current_file_num        9
static PyObject* fs_name_total_bytes;
#define fs_i_total_bytes             10
static PyObject* fs_name_total_transferred_bytes;
#define fs_i_total_transferred_bytes 11
static PyObject* fs_name_working_dir;
#define fs_i_working_dir             12
static PyObject* fs_name_current_file;
#define fs_i_current_file            13
static PyObject* fs_name_current_fd;
#define fs_i_current_fd              14
static PyObject* fs_name_current_file_size;
#define fs_i_current_file_size       15
static PyObject* fs_name_current_file_progress;
#define fs_i_current_file_progress   16
static PyObject* fs_name_current_speed;
#define fs_i_current_speed           17

static PyObject* fs_name_to_i;

static PyObject*
PyICQFileSession_New(icq_FileSession* pfile)
{
	PyICQFileSession* ret = PyObject_New(
		PyICQFileSession, &PyICQFileSessionType);
	if (ret) {
		ret->fs = pfile;
	}
	return (PyObject*)ret;
}

static void
icqlib_DelPyICQFileSession(PyICQFileSession* self)
{
	PyObject_Del(self);
}

static PyObject*
icqlib_PyICQFileSession_SetSpeed(PyICQFileSession* self, PyObject* args)
{
	int speed;

	if (!PyArg_ParseTuple(args,"i:SetSpeed",&speed)) {
		return NULL;
	}

	if (!self->fs) {
		PyErr_SetString(PyExc_RuntimeError,
			       "SetSpeed called on invalid FileSession");
		return NULL;
	}
	
	icq_FileSessionSetSpeed(self->fs,speed);
	
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject*
icqlib_PyICQFileSession_Close(PyICQFileSession* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":Close")) {
		return NULL;
	}

	if (!self->fs) {
		PyErr_SetString(PyExc_RuntimeError,
			       "Close called on invalid FileSession");
		return NULL;
	}
	
	icq_FileSessionClose(self->fs);

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject*
icqlib_PyICQFileSession_SetWorkingDir(PyICQFileSession* self, 
				      PyObject* args)
{
	char* dir;

	if (!PyArg_ParseTuple(args,"s:SetWorkingDir",&dir)) {
		return NULL;
	}

	if (!self->fs) {
		PyErr_SetString(PyExc_RuntimeError,
			       "SetWorkingDir called on invalid FileSession");
		return NULL;
	}
	
	icq_FileSessionSetWorkingDir(self->fs,dir);

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject*
icqlib_PyICQFileSession_SetFiles(PyICQFileSession* self, PyObject* args)
{
	PyObject* list;
	char** array;
	int i;

	if (!PyArg_ParseTuple(args,"O:SetFiles",&list)) {
		return NULL;
	}

	if (!self->fs) {
		PyErr_SetString(PyExc_RuntimeError,
			       "SetFiles called on invalid FileSession");
		return NULL;
	}	

	if (!PyList_Check(list)) {
		PyErr_Format(PyExc_TypeError,
			     "SetFiles: arg 1 must be a list, not a %.500s",
			     list->ob_type->tp_name);
		return NULL;
	}

	array = malloc(sizeof(char*) * (PyList_GET_SIZE(list)));
	
	if (!array) {
		PyErr_SetString(PyExc_MemoryError,
				"SetFiles: out of memory!");
		return NULL;
	}

	for (i = 0; i < PyList_GET_SIZE(list); i++) {
		array[i] = PyString_AsString(PyList_GET_ITEM(list,i));
		if (!array[i]) {
			free(array);
			return NULL;
		}
	}

	array[PyList_GET_SIZE(list)] = NULL;

	icq_FileSessionSetFiles(self->fs,array);

	free(array);

	Py_INCREF(Py_None);
	return Py_None;
}

#define PYICQFILESESSION_METH(NAME) \
  {#NAME, (PyCFunction)icqlib_PyICQFileSession_ ## NAME, METH_VARARGS \
   /*, icqlib_PyICQFileSession_ ## NAME ## __doc__*/}

static PyMethodDef icqlib_PyICQFileSession_methods[] = {
	PYICQFILESESSION_METH(SetSpeed),
	PYICQFILESESSION_METH(Close),
	PYICQFILESESSION_METH(SetWorkingDir),
	PYICQFILESESSION_METH(SetFiles),
	{NULL,NULL}
};

static PyObject*
icqlib_PyICQFileSession_getattro(PyICQFileSession* self, PyObject* attr)
{
	PyObject* v;

	v = PyDict_GetItem(fs_name_to_i, attr);

	if (!v) {
		char* s;

		s = PyString_AsString(attr);

		if (!s) return NULL;
		
#undef Py_FindMethod

		return Py_FindMethod(icqlib_PyICQFileSession_methods,
				     (PyObject*)self,s);
	} else {
		PyObject* w = NULL;
		icq_FileSession* fs = self->fs;
		switch (PyInt_AS_LONG(v)) {
		case fs_i_id:
			w = PyInt_FromLong(fs->id);
			break;

		case fs_i_status:
			w = PyInt_FromLong(fs->status);
			break;

		case fs_i_icqlink:
			w = (PyObject*)fs->icqlink->icq_UserData;
			break;

		case fs_i_direction:
			w = PyInt_FromLong(fs->direction);
			break;

		case fs_i_remote_uin:
			w = PyInt_FromLong(fs->remote_uin);
			break;

		case fs_i_remote_handle:
			w = PyString_FromString(fs->remote_handle);
			break;

		case fs_i_files: {
			int i;

			w = PyList_New(fs->total_files);

			for (i = 0; i < fs->total_files; i++) 
			    PyList_SET_ITEM(w,i,
				PyString_FromString(fs->files[i]));
			} break;

		case fs_i_total_files:
			w = PyInt_FromLong(fs->total_files);
			break;

		case fs_i_current_file_num:
			w = PyInt_FromLong(fs->current_file_num);
			break;

		case fs_i_total_bytes:
			w = PyInt_FromLong(fs->total_bytes);
			break;

		case fs_i_total_transferred_bytes:
			w = PyInt_FromLong(fs->total_transferred_bytes);
			break;

		case fs_i_working_dir:
			w = PyString_FromString(fs->working_dir);
			break;

		case fs_i_current_file:
			w = PyString_FromString(fs->current_file);
			break;

		case fs_i_current_fd:
			w = PyInt_FromLong(fs->current_fd);
			break;

		case fs_i_current_file_size:
			w = PyInt_FromLong(fs->current_file_size);
			break;

		case fs_i_current_file_progress:
			w = PyInt_FromLong(fs->current_file_progress);
			break;

		case fs_i_current_speed:
			w = PyInt_FromLong(fs->current_speed);
			break;
		}
		return w;
	}
}

static PyTypeObject PyICQFileSessionType = {
	PyObject_HEAD_INIT(NULL)
	0,
	"PyICQFileSession",
        sizeof(PyICQFileSession),      /*tp_basicsize*/
        0,                      /*tp_itemsize*/
        /* methods */
        (destructor)icqlib_DelPyICQFileSession, /*tp_dealloc*/
        0,                               /*tp_print*/
        0,                               /*tp_getattr*/
        0,                               /*tp_setattr*/
        0,                      /*tp_compare*/
        0,                      /*tp_repr*/
        0,                      /*tp_as_number*/
        0,                      /*tp_as_sequence*/
        0,                      /*tp_as_mapping*/
        0,                      /*tp_hash*/
	0, /*tp_call*/
	0, /*tp_str*/
	(getattrofunc)icqlib_PyICQFileSession_getattro, /*tp_getattro*/
	(setattrofunc)/*icqlib_PyICQLink_setattro*/0, /*tp_setattro*/
	0, /*tp_as_buffer*/
	0, /*tp_flags*/

	"" /*tp_doc*/
};

/***********************************************************
 *                    PyICQChatSession                     *
 ***********************************************************/

static PyObject* cs_name_status;
#define cs_i_status 0
static PyObject* cs_name_icqlink;
#define cs_i_icqlink 1
static PyObject* cs_name_remote_handle;
#define cs_i_remote_handle 2

static PyObject* cs_name_to_i;

static PyObject*
PyICQChatSession_New(icq_ChatSession* pchat)
{
	PyICQChatSession* ret = PyObject_New(
		PyICQChatSession, &PyICQChatSessionType);
	if (ret) {
		ret->cs = pchat;
	}
	return (PyObject*)ret;
}

void
PyICQChatSession_Del(PyICQChatSession* self)
{
	PyObject_Del(self);
}

static PyObject*
icqlib_PyICQChatSession_Close(PyICQChatSession* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":Close")) {
		return NULL;
	}

	if (!self->cs) {
		PyErr_SetString(PyExc_RuntimeError,
			       "Close called on invalid ChatSession");
		return NULL;
	}
	
	icq_ChatSessionClose(self->cs);

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject*
icqlib_PyICQChatSession_SendData(PyICQChatSession* self, PyObject* args)
{
	char* data;
	int length;

	if (!PyArg_ParseTuple(args,"s#:SendData",&data,&length)) {
		return NULL;
	}

	if (!self->cs) {
		PyErr_SetString(PyExc_RuntimeError,
			       "SendData called on invalid ChatSession");
		return NULL;
	}
	
	icq_ChatSessionSendData_n(self->cs,data,length);

	Py_INCREF(Py_None);
	return Py_None;
}

#define PYICQCHATSESSION_METH(NAME) \
  {#NAME, (PyCFunction)icqlib_PyICQChatSession_ ## NAME, METH_VARARGS \
   /*, icqlib_PyICQChatSession_ ## NAME ## __doc__*/}

static PyMethodDef icqlib_PyICQChatSession_methods[] = {
	PYICQCHATSESSION_METH(Close),
	PYICQCHATSESSION_METH(SendData),
	{NULL, NULL}
};

static PyObject*
icqlib_PyICQChatSession_getattro(PyICQChatSession* self, PyObject* attr)
{
	PyObject* v;

	v = PyDict_GetItem(fs_name_to_i, attr);

	if (!v) {
		char* s;

		s = PyString_AsString(attr);

		if (!s) return NULL;
		
		return Py_FindMethod(icqlib_PyICQChatSession_methods,
				     (PyObject*)self,s);
	}
	else {
		PyObject* w = NULL;
		icq_ChatSession* cs = self->cs;
		switch (PyInt_AS_LONG(v)) {
		case cs_i_status:
			w = PyInt_FromLong(cs->status);
			break;
		case cs_i_icqlink:
			w = (PyObject*)cs->icqlink->icq_UserData;
			break;
		case cs_i_remote_handle:
			w = PyString_FromString(cs->remote_handle);
			break;
		}
		return w;
	}
}
static PyTypeObject PyICQChatSessionType = {
	PyObject_HEAD_INIT(NULL)
	0,
	"PyICQChatSession",
        sizeof(PyICQChatSession),      /*tp_basicsize*/
        0,                      /*tp_itemsize*/
        /* methods */
        (destructor)PyICQChatSession_Del, /*tp_dealloc*/
        0,                               /*tp_print*/
        0,                               /*tp_getattr*/
        0,                               /*tp_setattr*/
        0,                      /*tp_compare*/
        0,                      /*tp_repr*/
        0,                      /*tp_as_number*/
        0,                      /*tp_as_sequence*/
        0,                      /*tp_as_mapping*/
        0,                      /*tp_hash*/
	0, /*tp_call*/
	0, /*tp_str*/
	(getattrofunc)icqlib_PyICQChatSession_getattro, /*tp_getattro*/
	(setattrofunc)/*icqlib_PyICQLink_setattro*/0, /*tp_setattro*/
	0, /*tp_as_buffer*/
	0, /*tp_flags*/

	"" /*tp_doc*/
};

/***********************************************************
 *                  module-level functions                 *
 ***********************************************************/

#define ICQLIB_DOC(FUNC) \
  static char icqlib_ ## FUNC ## __doc__[]


static PyObject*
icqlib_GetCountryName(PyObject* self, PyObject* args)
{
	unsigned short code;

	if (!PyArg_ParseTuple(args, "h:GetCountryName", &code)) {
		return NULL;
	}

	return PyString_FromString(icq_GetCountryName(code));
}

ICQLIB_DOC(GetCountryName) = 
"GetCountryName(code) -> string\n\
\n\
Convert the country code found in the meta user info to a more\n\
readable from.";

static PyObject*
icqlib_GetMetaOccupationName(PyObject* self, PyObject* args)
{
	unsigned short code;

	if (!PyArg_ParseTuple(args, "h:GetMetaOccupationName", &code)) {
		return NULL;
	}

	return PyString_FromString(icq_GetMetaOccupationName(code));
}

ICQLIB_DOC(GetMetaOccupationName) = 
"GetMetaOccupationName(code) -> string\n\
\n\
";

static PyObject*
icqlib_GetMetaBackgroundName(PyObject* self, PyObject* args)
{
	unsigned short code;

	if (!PyArg_ParseTuple(args, "h:GetMetaBackgroundName", &code)) {
		return NULL;
	}

	return PyString_FromString(icq_GetMetaBackgroundName(code));
}

ICQLIB_DOC(GetMetaBackgroundName) = 
"GetMetaBackgroundName(code) -> string\n\
\n\
";

static PyObject*
icqlib_GetMetaAffiliationName(PyObject* self, PyObject* args)
{
	unsigned short code;

	if (!PyArg_ParseTuple(args, "h:GetMetaAffiliationName", &code)) {
		return NULL;
	}

	return PyString_FromString(icq_GetMetaAffiliationName(code));
}

ICQLIB_DOC(GetMetaAffiliationName) = 
"GetMetaAffiliationName(code) -> string\n\
\n\
";

static PyObject*
icqlib_GetMetaLanguageName(PyObject* self, PyObject* args)
{
	unsigned short code;

	if (!PyArg_ParseTuple(args, "h:GetMetaLanguageName", &code)) {
		return NULL;
	}

	return PyString_FromString(icq_GetMetaLanguageName(code));
}

ICQLIB_DOC(GetMetaLanguageName) = 
"GetMetaLanguageName(code) -> string\n\
\n\
";

static PyObject* socketnotify;

static void my_socket_notify(int socket, int type, int status)
{
	PyObject* t;
	PyObject* v;

	t = Py_BuildValue("(iii)", socket, type, status);

	v = PyEval_CallObject(socketnotify, t);

	Py_DECREF(t);
	Py_XDECREF(v);
}

static PyObject*
icqlib_InstallSocketHandler(PyObject* self, PyObject* args)
{
	PyObject* callable;

	if (!PyArg_ParseTuple(args, "O:InstallSocketHandler",
			      &callable)) {
		return NULL;
	}

	Py_XDECREF(socketnotify);
	Py_INCREF(callable);
	socketnotify = callable;

	icq_SocketNotify = my_socket_notify;
		
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLIB_DOC(InstallSocketHandler) = 
"InstallSocketHandler(handler) -> None\n\
\n\
";

static PyObject* timeouthandler;

static void my_set_timeout(long seconds)
{
	PyObject* t;
	PyObject* v;

	t = Py_BuildValue("(i)", seconds);

	v = PyEval_CallObject(timeouthandler, t);

	Py_DECREF(t);
	Py_XDECREF(v);
}

static PyObject*
icqlib_InstallTimeoutHandler(PyObject* self, PyObject* args)
{
	PyObject* callable;

	if (!PyArg_ParseTuple(args, "O:InstallTimeoutHandler",
			      &callable)) {
		return NULL;
	}

	Py_XDECREF(timeouthandler);
	Py_INCREF(callable);
	timeouthandler = callable;

	icq_SetTimeout = my_set_timeout;
		
	Py_INCREF(Py_None);
	return Py_None;
}

ICQLIB_DOC(InstallTimeoutHandler) = 
"InstallTimeoutHandler(handler) -> None\n\
\n\
";

static PyObject*
icqlib_HandleReadySocket(PyObject* self, PyObject* args)
{
	int socket, type;

	if (!PyArg_ParseTuple(args, "ii:HandleReadySocket",
			      &socket, &type)) {
		return NULL;
	}

	icq_HandleReadySocket(socket, type);

	if (PyErr_Occurred()) {
		return NULL;
	} else {
		Py_INCREF(Py_None);
		return Py_None;
	}	
}

ICQLIB_DOC(HandleReadySocket) = 
"HandleReadySocket(fd, type)\n\
\n\
";

static PyObject*
icqlib_GetLogLevel(PyObject* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args, ":GetLogLevel")) {
		return NULL;
	}

	return PyInt_FromLong(icq_LogLevel);
}

ICQLIB_DOC(GetLogLevel) = 
"GetLogLevel() -> int\n\
\n\
";

static PyObject*
icqlib_SetLogLevel(PyObject* self, PyObject* args)
{
	int level;

	if (!PyArg_ParseTuple(args, "i:SetLogLevel",&level)) {
		return NULL;
	}

	icq_LogLevel = level;

	Py_INCREF(Py_None);
	return Py_None;
}

ICQLIB_DOC(SetLogLevel) = 
"SetLogLevel(level) -> None\n\
\n\
";

static PyObject*
icqlib_HandleTimeout(PyObject* self, PyObject* args)
{
	icq_HandleTimeout();

	if (PyErr_Occurred()) {
		return NULL;
	} else {
		Py_INCREF(Py_None);
		return Py_None;
	}	
}

ICQLIB_DOC(HandleTimeout) = 
"HandleTimeout() -> None\n\
\n\
";

static PyObject*
icqlib_Main(PyObject* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args,":Main")) {
		return NULL;
	}

	icq_Main();
	
	if (PyErr_Occurred()) {
		return NULL;
	} else {
		Py_INCREF(Py_None);
		return Py_None;
	}	
}

ICQLIB_DOC(Main) =
"Main() -> None\n\
\n\
Respond to a packet form the server, if there have been any.";

#define ICQLIB_FUNC(NAME) \
  { #NAME, (PyCFunction)icqlib_ ## NAME, METH_VARARGS \
  , icqlib_ ## NAME ## __doc__ }

static PyMethodDef icqlib_methods[] = {
	ICQLIB_FUNC(GetCountryName),
	ICQLIB_FUNC(GetMetaOccupationName),
	ICQLIB_FUNC(GetMetaBackgroundName),
	ICQLIB_FUNC(GetMetaAffiliationName),
	ICQLIB_FUNC(GetMetaLanguageName),
	ICQLIB_FUNC(InstallSocketHandler),
	ICQLIB_FUNC(InstallTimeoutHandler),
	ICQLIB_FUNC(HandleReadySocket),
	ICQLIB_FUNC(GetLogLevel),
	ICQLIB_FUNC(SetLogLevel),
	ICQLIB_FUNC(HandleTimeout),
	ICQLIB_FUNC(Main),
	{NULL,NULL}
};

static char icqlib___doc__[] = "\
A wrapper for icqlib.";

#define INTERN_NAME(NAME) \
  name_ ## NAME = PyString_InternFromString( #NAME )

#define INTERN_FS_NAME(NAME) \
  fs_name_ ## NAME = PyString_InternFromString( #NAME ); \
  do { PyObject* _int = PyInt_FromLong(fs_i_ ## NAME);  \
    if (_int) { \
      PyDict_SetItem(fs_name_to_i, fs_name_ ## NAME, \
                     PyInt_FromLong(fs_i_ ## NAME)); \
      Py_DECREF(_int); \
    } } while (0)

#define INTERN_CS_NAME(NAME) \
  cs_name_ ## NAME = PyString_InternFromString( #NAME ); \
  do { \
    PyObject* _int = PyInt_FromLong(cs_i_ ## NAME); \
    if (_int) { \
      PyDict_SetItem(cs_name_to_i, cs_name_ ## NAME, \
                     PyInt_FromLong(cs_i_ ## NAME)); \
      Py_DECREF(_int); \
    }\
  } while (0)


void
initicqlib(void)
{
	PyObject* m;
	PyObject* d;

	PyICQFileSessionType.ob_type = &PyType_Type;
	PyICQChatSessionType.ob_type = &PyType_Type;
	
	INTERN_NAME(Logged);
	INTERN_NAME(Disconnected);
	INTERN_NAME(RecvMessage);
	INTERN_NAME(RecvURL);
	INTERN_NAME(RecvWebPager);
	INTERN_NAME(RecvMailExpress);
	INTERN_NAME(RecvChatReq);
	INTERN_NAME(RecvFileReq);
	INTERN_NAME(RecvAdded);
	INTERN_NAME(RecvAuthReq);
	INTERN_NAME(UserFound);
	INTERN_NAME(SearchDone);
	INTERN_NAME(UserOnline);
	INTERN_NAME(UserOffline);
	INTERN_NAME(UserStatusUpdate);
	INTERN_NAME(InfoReply);
	INTERN_NAME(ExtInfoReply);
	INTERN_NAME(WrongPassword);
	INTERN_NAME(InvalidUIN);
	INTERN_NAME(Log);
	INTERN_NAME(SrvAck);
	INTERN_NAME(RequestNotify);
	INTERN_NAME(ChatNotify);
	INTERN_NAME(FileNotify);
	INTERN_NAME(NewUIN);
	INTERN_NAME(MetaUserFound);
	INTERN_NAME(MetaUserInfo);
	INTERN_NAME(MetaUserWork);
	INTERN_NAME(MetaUserMore);
	INTERN_NAME(MetaUserAbout);
	INTERN_NAME(MetaUserInterests);
	INTERN_NAME(MetaUserAffiliations);
	INTERN_NAME(MetaUserHomePageCategory);

	INTERN_NAME(Uin);
	INTERN_NAME(OurIP);
	INTERN_NAME(OurPort);
	INTERN_NAME(ContactList);
	INTERN_NAME(Status);
	INTERN_NAME(Password);
	INTERN_NAME(Nick);

	name_to_i = PyDict_New();

#define SET_NAME_TO_I(NAME) \
  PyDict_SetItem(name_to_i, name_ ## NAME, PyInt_FromLong( i_ ## NAME ))

	SET_NAME_TO_I(Uin);
	SET_NAME_TO_I(OurIP);
	SET_NAME_TO_I(OurPort);
/*	SET_NAME_TO_I(ContactList);*/
	SET_NAME_TO_I(Status);
	SET_NAME_TO_I(Password);
	SET_NAME_TO_I(Nick);

	fs_name_to_i = PyDict_New();

        INTERN_FS_NAME(id);
        INTERN_FS_NAME(status);
        INTERN_FS_NAME(icqlink);
        INTERN_FS_NAME(direction);
        INTERN_FS_NAME(remote_uin);
        INTERN_FS_NAME(remote_handle);
        INTERN_FS_NAME(files);
        INTERN_FS_NAME(total_files);
        INTERN_FS_NAME(current_file_num);
        INTERN_FS_NAME(total_bytes);
        INTERN_FS_NAME(total_transferred_bytes);
        INTERN_FS_NAME(working_dir);
        INTERN_FS_NAME(current_file);
        INTERN_FS_NAME(current_fd);
        INTERN_FS_NAME(current_file_size);
        INTERN_FS_NAME(current_file_progress);
        INTERN_FS_NAME(current_speed);

	cs_name_to_i = PyDict_New();

	INTERN_CS_NAME(status);
	INTERN_CS_NAME(icqlink);
	INTERN_CS_NAME(remote_handle);

	m = Py_InitModule3("icqlib", icqlib_methods, icqlib___doc__);
	d = PyModule_GetDict(m);

	PyExtensionClass_Export(d, "ICQLink", PyICQLinkType);
}
